@@ -6,21 +6,22 @@
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.*
+ * License as published by the Free Software Foundation and 
+ * appearing in the file LICENSE.LGPL included in the package of this file.
+ * You can also redistribute it and/or modify it under the terms of 
+ * the GNU General Public License as published by the Free Software Foundation and 
+ * appearing in the file LICENSE.GPL included in the package of this file.
+ *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
- * GNU Lesser General Public License for more details.*
- * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
- * Boston, MA  02111-1307  USA
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
 
+
 #include <errno.h>
 #include <string.h>
 
+#include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 
@@ -35,6 +36,8 @@
 #include "scim-bridge-agent-panel-client.h"
 #include "scim-bridge-output.h"
 
+using std::vector;
+
 using namespace scim;
 
 typedef Signal1<void, int> ScimBridgeAgentPanelClientSignalVoid;
@@ -67,7 +70,7 @@ class ScimBridgeAgentPanelClientImpl: public ScimBridgeAgentPanelClient
         void turn_off ();
         void update_screen ();
         void show_help (const String &help);
-        void show_factory_menu (const std::vector<PanelFactoryInfo> &menu);
+        void show_factory_menu (const vector<PanelFactoryInfo> &menu);
         void focus_in (const scim::String &uuid);
         void focus_out ();
         void update_factory_info (const PanelFactoryInfo &info);
@@ -114,9 +117,8 @@ class ScimBridgeAgentPanelClientImpl: public ScimBridgeAgentPanelClient
         SocketClient socket_client;
         int socket_timeout;
         uint32 socket_magic_key;
-        Transaction send_transaction;
-        scim_bridge_imcontext_id_t current_imcontext_id;
-        int send_refcount;
+        Transaction sending_transaction;
+        scim_bridge_imcontext_id_t prepared_imcontext_id;
 
         ScimBridgeAgentPanelClientSignalVoid  signal_reload_config;
         ScimBridgeAgentPanelClientSignalVoid signal_exit;
@@ -134,7 +136,7 @@ class ScimBridgeAgentPanelClientImpl: public ScimBridgeAgentPanelClient
         ScimBridgeAgentPanelClientSignalVoid signal_request_factory_menu;
         ScimBridgeAgentPanelClientSignalString signal_change_factory;
 
-        void launch_panel () const;
+        retval_t launch_panel ();
 
 };
 
@@ -156,7 +158,7 @@ ScimBridgeAgentPanelClient::~ScimBridgeAgentPanelClient ()
 
 
 ScimBridgeAgentPanelClientImpl::ScimBridgeAgentPanelClientImpl (const String &new_config_name, const ScimBridgeDisplay *new_display):
-config_name (new_config_name), display (NULL), socket_timeout (scim_get_default_socket_timeout ()), socket_magic_key (0), current_imcontext_id (-1), send_refcount (0)
+config_name (new_config_name), display (NULL), socket_timeout (scim_get_default_socket_timeout ()), socket_magic_key (0), prepared_imcontext_id (-1)
 {
     display = scim_bridge_alloc_display ();
     scim_bridge_copy_display (display, new_display);
@@ -165,20 +167,22 @@ config_name (new_config_name), display (NULL), socket_timeout (scim_get_default_
 
 ScimBridgeAgentPanelClientImpl::~ScimBridgeAgentPanelClientImpl ()
 {
+    if (socket_client.is_connected ()) close_connection ();
     scim_bridge_free_display (display);
 }
 
 
-void ScimBridgeAgentPanelClientImpl::launch_panel () const
+retval_t ScimBridgeAgentPanelClientImpl::launch_panel ()
 {
     scim_bridge_pdebugln (7, "launch_panel ()");
 
     String panel_program = scim_global_config_read (SCIM_GLOBAL_CONFIG_DEFAULT_PANEL_PROGRAM, String (SCIM_PANEL_PROGRAM));
-
-    if (panel_program.length () <= 0) panel_program = String (SCIM_PANEL_PROGRAM);
     if (panel_program [0] != SCIM_PATH_DELIM) panel_program = String (SCIM_LIBEXECDIR) + String (SCIM_PATH_DELIM_STRING) + panel_program;
 
-    if (access (panel_program.c_str (), X_OK) != 0) panel_program = String (SCIM_PANEL_PROGRAM);
+    struct stat stat_buf;
+    if (lstat (panel_program.c_str (), &stat_buf) || !S_ISREG (stat_buf.st_mode) || access (panel_program.c_str (), X_OK)) {
+        panel_program = String (SCIM_PANEL_PROGRAM);
+    }
     scim_bridge_pdebugln (5, "Launch \"%s\"", panel_program.c_str ());
 
     if (display != NULL) {
@@ -194,51 +198,44 @@ void ScimBridgeAgentPanelClientImpl::launch_panel () const
                 scim_bridge_perrorln ("Failed to forking for a panel process: %s", errno == 0 ? "Unknown reason":strerror (errno));
                 abort ();
             }
+            return RETVAL_FAILED;
         } else if (retval > 0) {
             waitpid (retval, NULL, 0);
+            return RETVAL_SUCCEEDED;
         } else {
             scim_bridge_perrorln ("Failed to launch a panel: %s", errno == 0 ? "Unknown reason":strerror (errno));
+            return RETVAL_FAILED;
         }
     } else {
-        scim_bridge_println ("Not yet supported!");
-        abort ();
+        scim_bridge_println ("No display for the panel");
+        return RETVAL_FAILED;
     }
 }
 
 
 int ScimBridgeAgentPanelClientImpl::open_connection ()
 {
-    SocketAddress address (scim_get_default_panel_socket_address (scim_bridge_display_get_name (display)));
     if (socket_client.is_connected ()) close_connection ();
+    SocketAddress address (scim_get_default_panel_socket_address (scim_bridge_display_get_name (display)));
 
-    bool retval;
-    int count = 0;
-
-    // Try three times.
-    while (true) {
-        if ((retval = socket_client.connect (address)) == false) {
-            usleep (100000);
-            launch_panel ();
-            for (int i = 0; i < 200; ++i) {
-                if (socket_client.connect (address)) {
-                    retval = true;
-                    break;
-                }
-                usleep (100000);
-            }
-        }
-
-        if (retval && scim_socket_open_connection (socket_magic_key, String ("FrontEnd"), String ("Panel"), socket_client, socket_timeout))
-            break;
-
-        socket_client.close ();
-
-        if (count++ >= 3) break;
-
+    if (!socket_client.connect (address)) {
+        launch_panel ();
         usleep (100000);
     }
+        
+    // Try to establish the connection six times.
+    for (int i = 0; i < 6; ++i) {
+        if (!socket_client.connect (address)) {
+            usleep (100000);
+        } else {
+            if (scim_socket_open_connection (socket_magic_key, String ("FrontEnd"), String ("Panel"), socket_client, socket_timeout)) {
+                return socket_client.get_id ();
+            }
+        }
+        socket_client.close ();
+    }
 
-    return socket_client.get_id ();
+    return -1;
 }
 
 
@@ -246,6 +243,7 @@ void ScimBridgeAgentPanelClientImpl::close_connection ()
 {
     socket_client.close ();
     socket_magic_key = 0;
+    prepared_imcontext_id = -1;
 }
 
 
@@ -269,115 +267,115 @@ bool ScimBridgeAgentPanelClientImpl::has_pending_event () const
 
 retval_t ScimBridgeAgentPanelClientImpl::filter_event ()
 {
-    Transaction recv_transaction;
+    Transaction receiving_transaction;
 
-    if (!socket_client.is_connected () || !recv_transaction.read_from_socket (socket_client, socket_timeout)) return RETVAL_FAILED;
+    if (!socket_client.is_connected () || !receiving_transaction.read_from_socket (socket_client, socket_timeout)) return RETVAL_FAILED;
 
-    int cmd;
-    uint32 context = static_cast<uint32> (-1);
+    int command;
+    if (!receiving_transaction.get_command (command) || command != SCIM_TRANS_CMD_REPLY) return RETVAL_SUCCEEDED;
 
-    if (!recv_transaction.get_command (cmd) || cmd != SCIM_TRANS_CMD_REPLY) return RETVAL_SUCCEEDED;
-
-    // No context id available, so there will be some global command.
-    if (recv_transaction.get_data_type () == SCIM_TRANS_DATA_COMMAND) {
-        while (recv_transaction.get_command (cmd)) {
-            switch (cmd) {
+    // No IMContext ID is available; this is a global command.
+    if (receiving_transaction.get_data_type () == SCIM_TRANS_DATA_COMMAND) {
+        while (receiving_transaction.get_command (command)) {
+            switch (command) {
                 case SCIM_TRANS_CMD_RELOAD_CONFIG:
-                    signal_reload_config (context);
+                    signal_reload_config (-1);
                     break;
                 case SCIM_TRANS_CMD_EXIT:
-                    signal_exit (context);
+                    signal_exit (-1);
                     break;
                 default:
                     break;
             }
         }
-        return RETVAL_SUCCEEDED;
-    }
 
-    // Now for context related command.
-    if (!recv_transaction.get_data (context)) return RETVAL_SUCCEEDED;
+    } else {
 
-    while (recv_transaction.get_command (cmd)) {
-        switch (cmd) {
-            case SCIM_TRANS_CMD_UPDATE_LOOKUP_TABLE_PAGE_SIZE:
-            {
-                uint32 size;
-                if (recv_transaction.get_data (size)) signal_update_lookup_table_page_size (context, (int) size);
-            }
-            break;
-            case SCIM_TRANS_CMD_LOOKUP_TABLE_PAGE_UP:
-            {
-                signal_lookup_table_page_up (context);
-            }
-            break;
-            case SCIM_TRANS_CMD_LOOKUP_TABLE_PAGE_DOWN:
-            {
-                signal_lookup_table_page_down (context);
-            }
-            break;
-            case SCIM_TRANS_CMD_TRIGGER_PROPERTY:
-            {
-                String property;
-                if (recv_transaction.get_data (property)) signal_trigger_property (context, property);
-            }
-            break;
-            case SCIM_TRANS_CMD_PROCESS_HELPER_EVENT:
-            {
-                String target_uuid;
-                String helper_uuid;
-                Transaction trans;
-                if (recv_transaction.get_data (target_uuid) && recv_transaction.get_data (helper_uuid) && recv_transaction.get_data (trans)) signal_process_helper_event (context, target_uuid, helper_uuid, trans);
-            }
-            break;
-            case SCIM_TRANS_CMD_MOVE_PREEDIT_CARET:
-            {
-                uint32 caret;
-                if (recv_transaction.get_data (caret)) signal_move_preedit_caret (context, caret);
-            }
-            break;
-            case SCIM_TRANS_CMD_SELECT_CANDIDATE:
-            {
-                uint32 item;
-                if (recv_transaction.get_data (item)) signal_select_candidate (context, item);
-            }
-            break;
-            case SCIM_TRANS_CMD_PROCESS_KEY_EVENT:
-            {
-                KeyEvent key;
-                if (recv_transaction.get_data (key)) signal_process_key_event (context, key);
-            }
-            break;
-            case SCIM_TRANS_CMD_COMMIT_STRING:
-            {
-                WideString wstr;
-                if (recv_transaction.get_data (wstr)) signal_commit_string (context, wstr);
-            }
-            break;
-            case SCIM_TRANS_CMD_FORWARD_KEY_EVENT:
-            {
-                KeyEvent key;
-                if (recv_transaction.get_data (key)) signal_forward_key_event (context, key);
-            }
-            break;
-            case SCIM_TRANS_CMD_PANEL_REQUEST_HELP:
-            {
-                signal_request_help (context);
-            }
-            break;
-            case SCIM_TRANS_CMD_PANEL_REQUEST_FACTORY_MENU:
-            {
-                signal_request_factory_menu (context);
-            }
-            break;
-            case SCIM_TRANS_CMD_PANEL_CHANGE_FACTORY:
-            {
-                String factory_uuid;
-                if (recv_transaction.get_data (factory_uuid)) signal_change_factory (context, factory_uuid);
-            }
-            break;
-            default:
+        // Check if there is a command for the specific IMContext.
+        uint32 imcontext_id;
+        if (!receiving_transaction.get_data (imcontext_id)) return RETVAL_SUCCEEDED;
+
+        while (receiving_transaction.get_command (command)) {
+            switch (command) {
+                case SCIM_TRANS_CMD_UPDATE_LOOKUP_TABLE_PAGE_SIZE:
+                {
+                    uint32 size;
+                    if (receiving_transaction.get_data (size)) signal_update_lookup_table_page_size (imcontext_id, (int) size);
+                }
                 break;
+                case SCIM_TRANS_CMD_LOOKUP_TABLE_PAGE_UP:
+                {
+                    signal_lookup_table_page_up (imcontext_id);
+                }
+                break;
+                case SCIM_TRANS_CMD_LOOKUP_TABLE_PAGE_DOWN:
+                {
+                    signal_lookup_table_page_down (imcontext_id);
+                }
+                break;
+                case SCIM_TRANS_CMD_TRIGGER_PROPERTY:
+                {
+                    String property;
+                    if (receiving_transaction.get_data (property)) signal_trigger_property (imcontext_id, property);
+                }
+                break;
+                case SCIM_TRANS_CMD_PROCESS_HELPER_EVENT:
+                {
+                    String target_uuid;
+                    String helper_uuid;
+                    Transaction trans;
+                    if (receiving_transaction.get_data (target_uuid) && receiving_transaction.get_data (helper_uuid) && receiving_transaction.get_data (trans)) signal_process_helper_event (imcontext_id, target_uuid, helper_uuid, trans);
+                }
+                break;
+                case SCIM_TRANS_CMD_MOVE_PREEDIT_CARET:
+                {
+                    uint32 cursor_pos;
+                    if (receiving_transaction.get_data (cursor_pos)) signal_move_preedit_caret (imcontext_id, cursor_pos);
+                }
+                break;
+                case SCIM_TRANS_CMD_SELECT_CANDIDATE:
+                {
+                    uint32 item;
+                    if (receiving_transaction.get_data (item)) signal_select_candidate (imcontext_id, item);
+                }
+                break;
+                case SCIM_TRANS_CMD_PROCESS_KEY_EVENT:
+                {
+                    KeyEvent key_event;
+                    if (receiving_transaction.get_data (key_event)) signal_process_key_event (imcontext_id, key_event);
+                }
+                break;
+                case SCIM_TRANS_CMD_FORWARD_KEY_EVENT:
+                {
+                    KeyEvent key_event;
+                    if (receiving_transaction.get_data (key_event)) signal_forward_key_event (imcontext_id, key_event);
+                }
+                break;
+                case SCIM_TRANS_CMD_COMMIT_STRING:
+                {
+                    WideString wstr;
+                    if (receiving_transaction.get_data (wstr)) signal_commit_string (imcontext_id, wstr);
+                }
+                break;
+                case SCIM_TRANS_CMD_PANEL_REQUEST_HELP:
+                {
+                    signal_request_help (imcontext_id);
+                }
+                break;
+                case SCIM_TRANS_CMD_PANEL_REQUEST_FACTORY_MENU:
+                {
+                    signal_request_factory_menu (imcontext_id);
+                }
+                break;
+                case SCIM_TRANS_CMD_PANEL_CHANGE_FACTORY:
+                {
+                    String factory_uuid;
+                    if (receiving_transaction.get_data (factory_uuid)) signal_change_factory (imcontext_id, factory_uuid);
+                }
+                break;
+                default:
+                    break;
+            }
         }
     }
 
@@ -387,29 +385,26 @@ retval_t ScimBridgeAgentPanelClientImpl::filter_event ()
 
 retval_t ScimBridgeAgentPanelClientImpl::prepare (scim_bridge_imcontext_id_t imcontext_id)
 {
-    if (!socket_client.is_connected ()) return RETVAL_FAILED;
-
-    int cmd;
-    uint32 data;
-    int new_imcontext_id = imcontext_id;
-
-    if (send_refcount <= 0) {
-        current_imcontext_id = imcontext_id;
-        send_transaction.clear ();
-        send_transaction.put_command (SCIM_TRANS_CMD_REQUEST);
-        send_transaction.put_data (socket_magic_key);
-        send_transaction.put_data ((uint32) new_imcontext_id);
-
-        send_transaction.get_command (cmd);
-        send_transaction.get_data (data);
-        send_transaction.get_data (data);
-        send_refcount = 0;
+    if (!socket_client.is_connected ()) {
+        if (open_connection ()) return RETVAL_FAILED;
     }
+    
+    if (prepared_imcontext_id == -1) {
+        sending_transaction.clear ();
+        sending_transaction.put_command (SCIM_TRANS_CMD_REQUEST);
+        sending_transaction.put_data (socket_magic_key);
+        sending_transaction.put_data ((uint32) imcontext_id);
 
-    if (new_imcontext_id == current_imcontext_id) {
-        ++send_refcount;
+        int command;
+        uint32 data;
+        sending_transaction.get_command (command);
+        sending_transaction.get_data (data);
+        sending_transaction.get_data (data);
+
+        prepared_imcontext_id = imcontext_id;
         return RETVAL_SUCCEEDED;
     } else {
+        scim_bridge_pdebugln (8, "The panel client is already prepared!");
         return RETVAL_FAILED;
     }
 }
@@ -419,14 +414,11 @@ retval_t ScimBridgeAgentPanelClientImpl::send ()
 {
     if (!socket_client.is_connected ()) return RETVAL_SUCCEEDED;
 
-    if (send_refcount <= 0) return RETVAL_SUCCEEDED;
-    --send_refcount;
+    if (prepared_imcontext_id == -1) return RETVAL_SUCCEEDED;
 
-    if (send_refcount > 0) return RETVAL_SUCCEEDED;
-
-    current_imcontext_id = -1;
-    if (send_transaction.get_data_type () != SCIM_TRANS_DATA_UNKNOWN) {
-        return send_transaction.write_to_socket (socket_client, 0x4d494353);
+    prepared_imcontext_id = -1;
+    if (sending_transaction.get_data_type () != SCIM_TRANS_DATA_UNKNOWN) {
+        return sending_transaction.write_to_socket (socket_client, 0x4d494353);
     } else {
         return RETVAL_SUCCEEDED;
     }
@@ -435,43 +427,43 @@ retval_t ScimBridgeAgentPanelClientImpl::send ()
 
 void ScimBridgeAgentPanelClientImpl::turn_on ()
 {
-    if (send_refcount > 0) send_transaction.put_command (SCIM_TRANS_CMD_PANEL_TURN_ON);
+    if (prepared_imcontext_id != -1) sending_transaction.put_command (SCIM_TRANS_CMD_PANEL_TURN_ON);
 }
 
 
 void ScimBridgeAgentPanelClientImpl::turn_off ()
 {
-    if (send_refcount > 0) send_transaction.put_command (SCIM_TRANS_CMD_PANEL_TURN_OFF);
+    if (prepared_imcontext_id != -1) sending_transaction.put_command (SCIM_TRANS_CMD_PANEL_TURN_OFF);
 }
 
 
 void ScimBridgeAgentPanelClientImpl::update_screen ()
 {
-    if (send_refcount > 0 && display != NULL) {
-        send_transaction.put_command (SCIM_TRANS_CMD_UPDATE_SCREEN);
-        send_transaction.put_data ((uint32) scim_bridge_display_get_screen_number (display));
+    if (prepared_imcontext_id != -1 && display != NULL) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_UPDATE_SCREEN);
+        sending_transaction.put_data ((uint32) scim_bridge_display_get_screen_number (display));
     }
 }
 
 
 void ScimBridgeAgentPanelClientImpl::show_help (const String &help)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_PANEL_SHOW_HELP);
-        send_transaction.put_data (help);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_PANEL_SHOW_HELP);
+        sending_transaction.put_data (help);
     }
 }
 
 
-void ScimBridgeAgentPanelClientImpl::show_factory_menu (const std::vector<PanelFactoryInfo> &menu)
+void ScimBridgeAgentPanelClientImpl::show_factory_menu (const vector<PanelFactoryInfo> &menu)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_PANEL_SHOW_FACTORY_MENU);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_PANEL_SHOW_FACTORY_MENU);
         for (size_t i = 0; i < menu.size (); ++i) {
-            send_transaction.put_data (menu[i].uuid);
-            send_transaction.put_data (menu[i].name);
-            send_transaction.put_data (menu[i].lang);
-            send_transaction.put_data (menu[i].icon);
+            sending_transaction.put_data (menu[i].uuid);
+            sending_transaction.put_data (menu[i].name);
+            sending_transaction.put_data (menu[i].lang);
+            sending_transaction.put_data (menu[i].icon);
         }
     }
 }
@@ -479,173 +471,173 @@ void ScimBridgeAgentPanelClientImpl::show_factory_menu (const std::vector<PanelF
 
 void ScimBridgeAgentPanelClientImpl::focus_in (const String &uuid)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_FOCUS_IN);
-        send_transaction.put_data (uuid);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_FOCUS_IN);
+        sending_transaction.put_data (uuid);
     }
 }
 
 
 void ScimBridgeAgentPanelClientImpl::focus_out ()
 {
-    if (send_refcount > 0) send_transaction.put_command (SCIM_TRANS_CMD_FOCUS_OUT);
+    if (prepared_imcontext_id != -1) sending_transaction.put_command (SCIM_TRANS_CMD_FOCUS_OUT);
 }
 
 
 void ScimBridgeAgentPanelClientImpl::update_factory_info (const PanelFactoryInfo &info)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_PANEL_UPDATE_FACTORY_INFO);
-        send_transaction.put_data (info.uuid);
-        send_transaction.put_data (info.name);
-        send_transaction.put_data (info.lang);
-        send_transaction.put_data (info.icon);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_PANEL_UPDATE_FACTORY_INFO);
+        sending_transaction.put_data (info.uuid);
+        sending_transaction.put_data (info.name);
+        sending_transaction.put_data (info.lang);
+        sending_transaction.put_data (info.icon);
     }
 }
 
 
 void ScimBridgeAgentPanelClientImpl::update_spot_location (int x, int y)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_UPDATE_SPOT_LOCATION);
-        send_transaction.put_data ((uint32) x);
-        send_transaction.put_data ((uint32) y);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_UPDATE_SPOT_LOCATION);
+        sending_transaction.put_data ((uint32) x);
+        sending_transaction.put_data ((uint32) y);
     }
 }
 
 
 void ScimBridgeAgentPanelClientImpl::show_preedit_string ()
 {
-    if (send_refcount > 0) send_transaction.put_command (SCIM_TRANS_CMD_SHOW_PREEDIT_STRING);
+    if (prepared_imcontext_id != -1) sending_transaction.put_command (SCIM_TRANS_CMD_SHOW_PREEDIT_STRING);
 }
 
 
 void ScimBridgeAgentPanelClientImpl::show_aux_string ()
 {
-    if (send_refcount > 0) send_transaction.put_command (SCIM_TRANS_CMD_SHOW_AUX_STRING);
+    if (prepared_imcontext_id != -1) sending_transaction.put_command (SCIM_TRANS_CMD_SHOW_AUX_STRING);
 }
 
 
 void ScimBridgeAgentPanelClientImpl::show_lookup_table ()
 {
-    if (send_refcount > 0) send_transaction.put_command (SCIM_TRANS_CMD_SHOW_LOOKUP_TABLE);
+    if (prepared_imcontext_id != -1) sending_transaction.put_command (SCIM_TRANS_CMD_SHOW_LOOKUP_TABLE);
 }
 
 
 void ScimBridgeAgentPanelClientImpl::hide_preedit_string ()
 {
-    if (send_refcount > 0) send_transaction.put_command (SCIM_TRANS_CMD_HIDE_PREEDIT_STRING);
+    if (prepared_imcontext_id != -1) sending_transaction.put_command (SCIM_TRANS_CMD_HIDE_PREEDIT_STRING);
 }
 
 
 void ScimBridgeAgentPanelClientImpl::hide_aux_string ()
 {
-    if (send_refcount > 0) send_transaction.put_command (SCIM_TRANS_CMD_HIDE_AUX_STRING);
+    if (prepared_imcontext_id != -1) sending_transaction.put_command (SCIM_TRANS_CMD_HIDE_AUX_STRING);
 }
 
 
 void ScimBridgeAgentPanelClientImpl::hide_lookup_table ()
 {
-    if (send_refcount > 0) send_transaction.put_command (SCIM_TRANS_CMD_HIDE_LOOKUP_TABLE);
+    if (prepared_imcontext_id != -1) sending_transaction.put_command (SCIM_TRANS_CMD_HIDE_LOOKUP_TABLE);
 }
 
 
 void ScimBridgeAgentPanelClientImpl::update_preedit_string (const WideString &str, const AttributeList &attrs)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_UPDATE_PREEDIT_STRING);
-        send_transaction.put_data (utf8_wcstombs (str));
-        send_transaction.put_data (attrs);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_UPDATE_PREEDIT_STRING);
+        sending_transaction.put_data (utf8_wcstombs (str));
+        sending_transaction.put_data (attrs);
     }
 }
 
 
 void ScimBridgeAgentPanelClientImpl::update_preedit_caret (int caret)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_UPDATE_PREEDIT_CARET);
-        send_transaction.put_data ((uint32) caret);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_UPDATE_PREEDIT_CARET);
+        sending_transaction.put_data ((uint32) caret);
     }
 }
 
 
 void ScimBridgeAgentPanelClientImpl::update_aux_string (const WideString &str, const AttributeList &attrs)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_UPDATE_AUX_STRING);
-        send_transaction.put_data (utf8_wcstombs (str));
-        send_transaction.put_data (attrs);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_UPDATE_AUX_STRING);
+        sending_transaction.put_data (utf8_wcstombs (str));
+        sending_transaction.put_data (attrs);
     }
 }
 
 
 void ScimBridgeAgentPanelClientImpl::update_lookup_table (const LookupTable &table)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_UPDATE_LOOKUP_TABLE);
-        send_transaction.put_data (table);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_UPDATE_LOOKUP_TABLE);
+        sending_transaction.put_data (table);
     }
 }
 
 
 void ScimBridgeAgentPanelClientImpl::register_properties (const PropertyList &properties)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_REGISTER_PROPERTIES);
-        send_transaction.put_data (properties);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_REGISTER_PROPERTIES);
+        sending_transaction.put_data (properties);
     }
 }
 
 
 void ScimBridgeAgentPanelClientImpl::update_property (const Property &property)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_UPDATE_PROPERTY);
-        send_transaction.put_data (property);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_UPDATE_PROPERTY);
+        sending_transaction.put_data (property);
     }
 }
 
 
 void ScimBridgeAgentPanelClientImpl::start_helper (const String &helper_uuid)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_START_HELPER);
-        send_transaction.put_data (helper_uuid);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_START_HELPER);
+        sending_transaction.put_data (helper_uuid);
     }
 }
 
 
 void ScimBridgeAgentPanelClientImpl::stop_helper (const String &helper_uuid)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_STOP_HELPER);
-        send_transaction.put_data (helper_uuid);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_STOP_HELPER);
+        sending_transaction.put_data (helper_uuid);
     }
 }
 
 
 void ScimBridgeAgentPanelClientImpl::send_helper_event (const String &helper_uuid, const Transaction &trans)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_SEND_HELPER_EVENT);
-        send_transaction.put_data (helper_uuid);
-        send_transaction.put_data (trans);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_SEND_HELPER_EVENT);
+        sending_transaction.put_data (helper_uuid);
+        sending_transaction.put_data (trans);
     }
 }
 
 
 void ScimBridgeAgentPanelClientImpl::register_input_context (const String &uuid)
 {
-    if (send_refcount > 0) {
-        send_transaction.put_command (SCIM_TRANS_CMD_PANEL_REGISTER_INPUT_CONTEXT);
-        send_transaction.put_data (uuid);
+    if (prepared_imcontext_id != -1) {
+        sending_transaction.put_command (SCIM_TRANS_CMD_PANEL_REGISTER_INPUT_CONTEXT);
+        sending_transaction.put_data (uuid);
     }
 }
 
 
 void ScimBridgeAgentPanelClientImpl::remove_input_context ()
 {
-    if (send_refcount > 0) send_transaction.put_command (SCIM_TRANS_CMD_PANEL_REMOVE_INPUT_CONTEXT);
+    if (prepared_imcontext_id != -1) sending_transaction.put_command (SCIM_TRANS_CMD_PANEL_REMOVE_INPUT_CONTEXT);
 }
 
 
