@@ -115,6 +115,29 @@ static retval_t received_message_unknown (const ScimBridgeMessage *message)
 }
 
 
+static retval_t received_message_imengine_status_changed (const ScimBridgeMessage *message)
+{
+    const char *header = scim_bridge_message_get_header (message);
+    const char *ic_id_str = scim_bridge_message_get_argument (message, 0);
+    const char *enabled_str = scim_bridge_message_get_argument (message, 1);
+
+    scim_bridge_imcontext_id_t ic_id;
+    boolean enabled;
+    if (scim_bridge_string_to_int (&ic_id, ic_id_str) || scim_bridge_string_to_boolean (&enabled, enabled_str)) {
+        scim_bridge_perrorln ("Invalid arguments for the message: %s (%s, %s)", header, ic_id_str, enabled_str);
+    } else {
+        ScimBridgeClientIMContext *imcontext = scim_bridge_client_find_imcontext (ic_id);
+        if (imcontext != NULL) {
+            scim_bridge_client_imcontext_imengine_status_changed (imcontext, enabled);
+        } else {
+            scim_bridge_perrorln ("No such imcontext: id = %d", ic_id);
+        }
+    }
+
+    return RETVAL_SUCCEEDED;
+}
+
+
 static retval_t received_message_preedit_mode_changed (const ScimBridgeMessage *message)
 {
     const char *header = scim_bridge_message_get_header (message);
@@ -787,11 +810,12 @@ static retval_t received_message_get_surrounding_text (const ScimBridgeMessage *
         if (imcontext != NULL) {
             succeeded = scim_bridge_client_imcontext_get_surrounding_text (imcontext, before_max, after_max, &string, &cursor_position);
             if (succeeded) {
-                scim_bridge_pdebugln (3, "surrounding text = '%s', cursor_position = %d", string, cursor_position);
+                scim_bridge_pdebugln (5, "surrounding text = '%s', cursor_position = %d", string, cursor_position);
             } else {
                 scim_bridge_pdebugln (5, "surrounding text = N/A");
             }
         } else {
+    
             scim_bridge_perrorln ("No such imcontext: id = %d", ic_id);
         }
     }
@@ -838,11 +862,11 @@ static retval_t received_message_delete_surrounding_text (const ScimBridgeMessag
     const char *length_str = scim_bridge_message_get_argument (message, 2);
 
     int ic_id;
-    unsigned int offset;
+    int offset;
     unsigned int length;
 
     boolean succeeded = FALSE;
-    if (scim_bridge_string_to_int (&ic_id, ic_id_str) || scim_bridge_string_to_uint (&offset, offset_str) || scim_bridge_string_to_uint (&length, length_str)) {
+    if (scim_bridge_string_to_int (&ic_id, ic_id_str) || scim_bridge_string_to_int (&offset, offset_str) || scim_bridge_string_to_uint (&length, length_str)) {
         scim_bridge_perrorln ("Invalid arguments for the message: %s (%s, %s, %s)", header, ic_id_str, offset_str, length_str);
     } else {
         ScimBridgeClientIMContext *imcontext = scim_bridge_client_find_imcontext (ic_id);
@@ -1184,7 +1208,9 @@ retval_t scim_bridge_client_read_and_dispatch ()
         scim_bridge_pdebugln (5, "");
 
         retval_t retval = RETVAL_SUCCEEDED;
-        if (strcmp (message_header, SCIM_BRIDGE_MESSAGE_PREEDIT_MODE_CHANGED) == 0) {
+        if (strcmp (message_header, SCIM_BRIDGE_MESSAGE_IMENGINE_STATUS_CHANGED) == 0) {
+            retval = received_message_imengine_status_changed (message);
+        } else if (strcmp (message_header, SCIM_BRIDGE_MESSAGE_PREEDIT_MODE_CHANGED) == 0) {
             retval = received_message_preedit_mode_changed (message);
         } else if (strcmp (message_header, SCIM_BRIDGE_MESSAGE_SET_COMMIT_STRING) == 0) {
             retval = received_message_set_commit_string (message);
