@@ -46,10 +46,8 @@ struct _ScimBridgeClientIMContext
 
     char *preedit_string;
     size_t preedit_string_capacity;
-
-    ScimBridgeAttribute **preedit_attributes;
-    size_t preedit_attribute_count;
-    size_t preedit_attribute_capacity;
+    
+    PangoAttrList *preedit_attributes;
 
     unsigned int preedit_cursor_position;
     boolean preedit_cursor_flicking;
@@ -59,6 +57,8 @@ struct _ScimBridgeClientIMContext
 
     char *commit_string;
     size_t commit_string_capacity;
+    
+    boolean enabled;
 
     GdkWindow *client_window;
 
@@ -83,6 +83,7 @@ static GtkWidget *focused_widget = NULL;
 
 static GtkIMContext *fallback_imcontext = NULL;
 static gulong fallback_commit_handler;
+static gulong fallback_preedit_changed_handler;
 
 static guint key_snooper_id = 0;
 static boolean key_snooper_used = FALSE;
@@ -108,7 +109,38 @@ static void scim_bridge_client_imcontext_set_cursor_location (GtkIMContext *cont
 static void fallback_commit (GtkIMContext *context, const char *str, gpointer data)
 {
     scim_bridge_pdebugln (4, "fallback_commit ()");
-    if (str != NULL && focused_imcontext != NULL) g_signal_emit_by_name (focused_imcontext, "commit", str);
+    if (focused_imcontext != NULL && !focused_imcontext->enabled && str != NULL) {
+        g_signal_emit_by_name (focused_imcontext, "commit", str);
+    }
+}
+
+static void fallback_preedit_changed (GtkIMContext *context, gpointer data)
+{
+    scim_bridge_pdebugln (4, "fallback_preedit_changed ()");
+    if (focused_imcontext != NULL && !focused_imcontext->enabled && context != NULL) {
+        gchar* preedit_string = NULL;
+        gint preedit_cursor_position = 0;
+        PangoAttrList *preedit_attributes = NULL;
+        gtk_im_context_get_preedit_string (context, &preedit_string, &preedit_attributes, &preedit_cursor_position);
+        
+        if (preedit_string != NULL) {
+            free (focused_imcontext->preedit_string);
+            focused_imcontext->preedit_string = preedit_string;
+            focused_imcontext->preedit_string_capacity = strlen (preedit_string);
+            focused_imcontext->preedit_shown = TRUE;
+        } else {
+            focused_imcontext->preedit_string[0] = '\0';
+            focused_imcontext->preedit_shown = FALSE;
+        }
+        
+        focused_imcontext->preedit_cursor_position = preedit_cursor_position;
+        
+        if (focused_imcontext->preedit_attributes != NULL)
+            pango_attr_list_unref (focused_imcontext->preedit_attributes);
+        focused_imcontext->preedit_attributes = preedit_attributes;
+        
+        g_signal_emit_by_name (focused_imcontext, "preedit-changed");
+    }
 }
 
 
@@ -189,7 +221,8 @@ static gboolean key_snooper (GtkWidget *widget, GdkEventKey *event, gpointer dat
             scim_bridge_perrorln ("An IOException at key_snooper ()");
             return FALSE;
         } else {
-            if (consumed) return TRUE;
+            if (consumed)
+                return TRUE;
         }
     }
 
@@ -200,7 +233,7 @@ static gboolean key_snooper (GtkWidget *widget, GdkEventKey *event, gpointer dat
 static boolean is_key_snooper_enabled ()
 {
     static boolean first_time = TRUE;
-    static boolean key_snooper_enabled = FALSE;
+    static boolean key_snooper_enabled = TRUE;
 
     if (first_time) {
         char *env_key_snooper_enabled = getenv ("SCIM_BRIDGE_KEY_SNOOPER_ENABLED");
@@ -270,8 +303,11 @@ void scim_bridge_client_imcontext_set_preedit_shown (ScimBridgeClientIMContext *
         imcontext->preedit_string = malloc (sizeof (char));
         imcontext->preedit_string[0] = '\0';
         imcontext->preedit_string_capacity = 0;
-        imcontext->preedit_attribute_count = 0;
         imcontext->preedit_cursor_position = 0;
+        if (imcontext->preedit_attributes != NULL) {
+            pango_attr_list_unref (imcontext->preedit_attributes);
+            imcontext->preedit_attributes = NULL;
+        }
     }
 }
 
@@ -282,22 +318,117 @@ void scim_bridge_client_imcontext_set_preedit_cursor_position (ScimBridgeClientI
 
 
 void scim_bridge_client_imcontext_set_preedit_attributes (ScimBridgeClientIMContext *imcontext, ScimBridgeAttribute** const preedit_attributes, int attribute_count)
-{
-    if (attribute_count > imcontext->preedit_attribute_capacity) {
-        imcontext->preedit_attributes = realloc (imcontext->preedit_attributes, sizeof (ScimBridgeAttribute*) * attribute_count);
-
-        int i;
-        for (i = imcontext->preedit_attribute_capacity; i < attribute_count; ++i) {
-            imcontext->preedit_attributes[i] = scim_bridge_alloc_attribute ();
-        }
-        imcontext->preedit_attribute_capacity = attribute_count;
+{   
+    if (imcontext->preedit_attributes != NULL)
+        pango_attr_list_unref (imcontext->preedit_attributes);
+    
+    imcontext->preedit_attributes = pango_attr_list_new ();
+    
+    int preedit_string_length = 0;
+    int preedit_wstring_length = 0;
+    if (imcontext->preedit_string != NULL) {
+        preedit_string_length = strlen (imcontext->preedit_string);
+        preedit_wstring_length = g_utf8_strlen (imcontext->preedit_string, -1);
     }
-
+    
+    boolean *has_attribute = alloca (sizeof (boolean) * preedit_string_length);
     int i;
+    for (i = 0; i < preedit_string_length; ++i) {
+        has_attribute[i] = FALSE;
+    }
+
     for (i = 0; i < attribute_count; ++i) {
-        scim_bridge_copy_attribute (imcontext->preedit_attributes[i], preedit_attributes[i]);
+        const ScimBridgeAttribute *attr = preedit_attributes[i];
+        const int begin_pos = scim_bridge_attribute_get_begin (attr);
+        const int end_pos = scim_bridge_attribute_get_end (attr);
+
+        if (begin_pos <= end_pos && 0 <= begin_pos && end_pos <= preedit_wstring_length) {
+            const int start_index = g_utf8_offset_to_pointer (imcontext->preedit_string, begin_pos) - imcontext->preedit_string;
+            const int end_index = g_utf8_offset_to_pointer (imcontext->preedit_string, end_pos) - imcontext->preedit_string;
+
+            const scim_bridge_attribute_type_t attr_type = scim_bridge_attribute_get_type (attr);
+            const scim_bridge_attribute_value_t attr_value = scim_bridge_attribute_get_value (attr);
+
+            boolean valid_attribute = FALSE;
+            if (attr_type == ATTRIBUTE_DECORATE) {
+                if (attr_value == SCIM_BRIDGE_ATTRIBUTE_DECORATE_UNDERLINE) {
+                    valid_attribute = TRUE;
+
+                    PangoAttribute *pango_attr = pango_attr_underline_new (PANGO_UNDERLINE_SINGLE);
+                    pango_attr->start_index = start_index;
+                    pango_attr->end_index = end_index;
+                    pango_attr_list_insert (imcontext->preedit_attributes, pango_attr);
+                } else if (attr_value == SCIM_BRIDGE_ATTRIBUTE_DECORATE_REVERSE) {
+                    valid_attribute = TRUE;
+
+                    PangoAttribute *pango_attr0 = pango_attr_foreground_new (preedit_normal_background.red, preedit_normal_background.green, preedit_normal_background.blue);
+                    pango_attr0->start_index = start_index;
+                    pango_attr0->end_index = end_index;
+                    pango_attr_list_insert (imcontext->preedit_attributes, pango_attr0);
+
+                    PangoAttribute *pango_attr1 = pango_attr_background_new (preedit_normal_foreground.red, preedit_normal_foreground.green, preedit_normal_foreground.blue);
+                    pango_attr1->start_index = start_index;
+                    pango_attr1->end_index = end_index;
+                    pango_attr_list_insert (imcontext->preedit_attributes, pango_attr1);
+                } else if (attr_value == SCIM_BRIDGE_ATTRIBUTE_DECORATE_HIGHLIGHT) {
+                    valid_attribute = TRUE;
+
+                    PangoAttribute *pango_attr0 = pango_attr_foreground_new (preedit_active_foreground.red, preedit_active_foreground.green, preedit_active_foreground.blue);
+                    pango_attr0->start_index = start_index;
+                    pango_attr0->end_index = end_index;
+                    pango_attr_list_insert (imcontext->preedit_attributes, pango_attr0);
+
+                    PangoAttribute *pango_attr1 = pango_attr_background_new (preedit_active_background.red, preedit_active_background.green, preedit_active_background.blue);
+                    pango_attr1->start_index = start_index;
+                    pango_attr1->end_index = end_index;
+                    pango_attr_list_insert (imcontext->preedit_attributes, pango_attr1);
+                } else {
+                    scim_bridge_perrorln ("Unknown preedit decoration!");
+                }
+            } else if (attr_type == ATTRIBUTE_FOREGROUND) {
+                valid_attribute = TRUE;
+
+                const unsigned int red = scim_bridge_attribute_get_red (attr) * 256;
+                const unsigned int green = scim_bridge_attribute_get_green (attr) * 256;
+                const unsigned int blue = scim_bridge_attribute_get_blue (attr) * 256;
+
+                PangoAttribute *pango_attr = pango_attr_foreground_new (red, green, blue);
+                pango_attr->start_index = start_index;
+                pango_attr->end_index = end_index;
+                pango_attr_list_insert (imcontext->preedit_attributes, pango_attr);
+            } else if (attr_type == ATTRIBUTE_BACKGROUND) {
+                valid_attribute = TRUE;
+
+                const unsigned int red = scim_bridge_attribute_get_red (attr) * 256;
+                const unsigned int green = scim_bridge_attribute_get_green (attr) * 256;
+                const unsigned int blue = scim_bridge_attribute_get_blue (attr) * 256;
+
+                PangoAttribute *pango_attr = pango_attr_background_new (red, green, blue);
+                pango_attr->start_index = start_index;
+                pango_attr->end_index = end_index;
+                pango_attr_list_insert (imcontext->preedit_attributes, pango_attr);
+            }
+
+            if (valid_attribute) {
+                int j;
+                for (j = start_index; j < end_index; ++j) {
+                    has_attribute[j] = TRUE;
+                }
+            }
+        }
+
+    }
+
+    // Add underlines for the all characters without attributes.
+    for (i = 0; i < preedit_string_length; ++i) {
+        if (has_attribute[i] == FALSE) {
+            PangoAttribute *pango_attr = pango_attr_underline_new (PANGO_UNDERLINE_SINGLE);
+            pango_attr->start_index = i;
+            for (; i < preedit_string_length && has_attribute[i] == FALSE; ++i);
+            pango_attr->end_index = i;
+            pango_attr_list_insert (imcontext->preedit_attributes, pango_attr);
+        }
     }
-    imcontext->preedit_attribute_count = attribute_count;
 }
 
 
@@ -444,6 +575,20 @@ void scim_bridge_client_imcontext_forward_key_event (ScimBridgeClientIMContext *
 }
 
 
+void scim_bridge_client_imcontext_imengine_status_changed (ScimBridgeClientIMContext *imcontext, boolean enabled)
+{
+    if (imcontext->preedit_shown) {
+        if (imcontext->enabled) {
+            scim_bridge_client_imcontext_set_preedit_shown (imcontext, FALSE);
+            scim_bridge_client_imcontext_update_preedit (imcontext);
+        } else {
+            gtk_im_context_reset (GTK_IM_CONTEXT (fallback_imcontext));
+        }
+    }
+    imcontext->enabled = enabled;
+}
+
+
 /* Class Implementations */
 void scim_bridge_client_imcontext_static_initialize ()
 {
@@ -456,6 +601,7 @@ void scim_bridge_client_imcontext_static_initialize ()
 
     fallback_imcontext = gtk_im_context_simple_new ();
     fallback_commit_handler = g_signal_connect (G_OBJECT (fallback_imcontext), "commit", G_CALLBACK (fallback_commit), NULL);
+    fallback_preedit_changed_handler = g_signal_connect (G_OBJECT (fallback_imcontext), "preedit_changed", G_CALLBACK (fallback_preedit_changed), NULL);
 }
 
 
@@ -560,13 +706,13 @@ void scim_bridge_client_imcontext_initialize (ScimBridgeClientIMContext *imconte
     imcontext->preedit_string_capacity = 0;
 
     imcontext->preedit_attributes = NULL;
-    imcontext->preedit_attribute_count = 0;
-    imcontext->preedit_attribute_capacity = 0;
 
     imcontext->commit_string = malloc (sizeof (char));
     imcontext->commit_string[0] = '\0';
     imcontext->commit_string_capacity = 0;
 
+    imcontext->enabled = FALSE;
+
     imcontext->client_window = NULL;
 
     imcontext->id = -1;
@@ -602,15 +748,11 @@ void scim_bridge_client_imcontext_finalize (GObject *object)
     free (imcontext->preedit_string);
     free (imcontext->commit_string);
 
-    int i;
-    for (i = 0; i < imcontext->preedit_attribute_capacity; ++i) {
-        scim_bridge_free_attribute (imcontext->preedit_attributes[i]);
-    }
-    free (imcontext->preedit_attributes);
+    if (imcontext->preedit_attributes != NULL)
+        pango_attr_list_unref (imcontext->preedit_attributes);
+    
     imcontext->preedit_attributes = NULL;
-    imcontext->preedit_attribute_count = 0;
-    imcontext->preedit_attribute_capacity = 0;
-
+    
     root_klass->finalize (object);
 }
 
@@ -652,7 +794,21 @@ gboolean scim_bridge_client_imcontext_filter_key_event (GtkIMContext *context, G
         }
     }
 
-    return gtk_im_context_filter_keypress (fallback_imcontext, event);
+    unsigned int accelerator_mask = (gtk_accelerator_get_default_mod_mask () & ~GDK_SHIFT_MASK);
+    if (imcontext == NULL || !imcontext->enabled) {
+        return gtk_im_context_filter_keypress (fallback_imcontext, event);
+    } else if (event->type == GDK_KEY_PRESS && (event->state & accelerator_mask) == 0) {
+        guint32 wchar = gdk_keyval_to_unicode (event->keyval);
+        if (wchar != 0) {
+            gchar buffer[10];
+            const int buffer_length = g_unichar_to_utf8 (wchar, buffer);
+            buffer[buffer_length] = '\0';
+            g_signal_emit_by_name (focused_imcontext, "commit", &buffer);
+            return TRUE;
+        }
+    }
+    
+    return FALSE;
 }
 
 
@@ -699,106 +855,8 @@ void scim_bridge_client_imcontext_get_preedit_string (GtkIMContext *context, gch
         }
         
         if (pango_attrs) {
-            *pango_attrs = pango_attr_list_new ();
-            
-            boolean *has_attribute = alloca (sizeof (boolean) * preedit_string_length);
-            int i;
-            for (i = 0; i < preedit_string_length; ++i) {
-                has_attribute[i] = FALSE;
-            }
-
-            for (i = 0; i < imcontext->preedit_attribute_count; ++i) {
-                const ScimBridgeAttribute *attr = imcontext->preedit_attributes[i];
-                const int begin_pos = scim_bridge_attribute_get_begin (attr);
-                const int end_pos = scim_bridge_attribute_get_end (attr);
-
-                if (begin_pos <= end_pos && 0 <= begin_pos && end_pos <= preedit_wstring_length) {
-                    const int start_index = g_utf8_offset_to_pointer (imcontext->preedit_string, begin_pos) - imcontext->preedit_string;
-                    const int end_index = g_utf8_offset_to_pointer (imcontext->preedit_string, end_pos) - imcontext->preedit_string;
-
-                    const scim_bridge_attribute_type_t attr_type = scim_bridge_attribute_get_type (attr);
-                    const scim_bridge_attribute_value_t attr_value = scim_bridge_attribute_get_value (attr);
-
-                    boolean valid_attribute = FALSE;
-                    if (attr_type == ATTRIBUTE_DECORATE) {
-                        if (attr_value == SCIM_BRIDGE_ATTRIBUTE_DECORATE_UNDERLINE) {
-                            valid_attribute = TRUE;
-
-                            PangoAttribute *pango_attr = pango_attr_underline_new (PANGO_UNDERLINE_SINGLE);
-                            pango_attr->start_index = start_index;
-                            pango_attr->end_index = end_index;
-                            pango_attr_list_insert (*pango_attrs, pango_attr);
-                        } else if (attr_value == SCIM_BRIDGE_ATTRIBUTE_DECORATE_REVERSE) {
-                            valid_attribute = TRUE;
-
-                            PangoAttribute *pango_attr0 = pango_attr_foreground_new (preedit_normal_background.red, preedit_normal_background.green, preedit_normal_background.blue);
-                            pango_attr0->start_index = start_index;
-                            pango_attr0->end_index = end_index;
-                            pango_attr_list_insert (*pango_attrs, pango_attr0);
-
-                            PangoAttribute *pango_attr1 = pango_attr_background_new (preedit_normal_foreground.red, preedit_normal_foreground.green, preedit_normal_foreground.blue);
-                            pango_attr1->start_index = start_index;
-                            pango_attr1->end_index = end_index;
-                            pango_attr_list_insert (*pango_attrs, pango_attr1);
-                        } else if (attr_value == SCIM_BRIDGE_ATTRIBUTE_DECORATE_HIGHLIGHT) {
-                            valid_attribute = TRUE;
-
-                            PangoAttribute *pango_attr0 = pango_attr_foreground_new (preedit_active_foreground.red, preedit_active_foreground.green, preedit_active_foreground.blue);
-                            pango_attr0->start_index = start_index;
-                            pango_attr0->end_index = end_index;
-                            pango_attr_list_insert (*pango_attrs, pango_attr0);
-
-                            PangoAttribute *pango_attr1 = pango_attr_background_new (preedit_active_background.red, preedit_active_background.green, preedit_active_background.blue);
-                            pango_attr1->start_index = start_index;
-                            pango_attr1->end_index = end_index;
-                            pango_attr_list_insert (*pango_attrs, pango_attr1);
-                        } else {
-                            scim_bridge_perrorln ("Unknown preedit decoration!");
-                        }
-                    } else if (attr_type == ATTRIBUTE_FOREGROUND) {
-                        valid_attribute = TRUE;
-
-                        const unsigned int red = scim_bridge_attribute_get_red (attr) * 256;
-                        const unsigned int green = scim_bridge_attribute_get_green (attr) * 256;
-                        const unsigned int blue = scim_bridge_attribute_get_blue (attr) * 256;
-
-                        PangoAttribute *pango_attr = pango_attr_foreground_new (red, green, blue);
-                        pango_attr->start_index = start_index;
-                        pango_attr->end_index = end_index;
-                        pango_attr_list_insert (*pango_attrs, pango_attr);
-                    } else if (attr_type == ATTRIBUTE_BACKGROUND) {
-                        valid_attribute = TRUE;
-
-                        const unsigned int red = scim_bridge_attribute_get_red (attr) * 256;
-                        const unsigned int green = scim_bridge_attribute_get_green (attr) * 256;
-                        const unsigned int blue = scim_bridge_attribute_get_blue (attr) * 256;
-
-                        PangoAttribute *pango_attr = pango_attr_background_new (red, green, blue);
-                        pango_attr->start_index = start_index;
-                        pango_attr->end_index = end_index;
-                        pango_attr_list_insert (*pango_attrs, pango_attr);
-                    }
-
-                    if (valid_attribute) {
-                        int j;
-                        for (j = start_index; j < end_index; ++j) {
-                            has_attribute[j] = TRUE;
-                        }
-                    }
-                }
-
-            }
-
-            // Add underlines for the all characters without attributes.
-            for (i = 0; i < preedit_string_length; ++i) {
-                if (has_attribute[i] == FALSE) {
-                    PangoAttribute *pango_attr = pango_attr_underline_new (PANGO_UNDERLINE_SINGLE);
-                    pango_attr->start_index = i;
-                    for (; i < preedit_string_length && has_attribute[i] == FALSE; ++i);
-                    pango_attr->end_index = i;
-                    pango_attr_list_insert (*pango_attrs, pango_attr);
-                }
-            }
+            *pango_attrs = imcontext->preedit_attributes;
+            pango_attr_list_ref (imcontext->preedit_attributes);
         }
     } else {
         if (str) *str = g_strdup ("");
@@ -843,11 +901,14 @@ void scim_bridge_client_imcontext_focus_out (GtkIMContext *context)
 
     focused_imcontext = imcontext;
     if (imcontext->preedit_shown) {
-        scim_bridge_client_imcontext_set_preedit_shown (imcontext, FALSE);
-        scim_bridge_client_imcontext_update_preedit (imcontext);
+        if (imcontext->enabled) {
+            scim_bridge_client_imcontext_set_preedit_shown (imcontext, FALSE);
+            scim_bridge_client_imcontext_update_preedit (imcontext);
+        } else {
+            gtk_im_context_reset (GTK_IM_CONTEXT (fallback_imcontext));
+        }
     }
     if (scim_bridge_client_is_messenger_opened () && imcontext != NULL) {
-
         if (scim_bridge_client_change_focus (imcontext, FALSE)) {
             scim_bridge_perrorln ("An IOException occurred at scim_bridge_client_imcontext_focus_out ()");
         }
