@@ -22,10 +22,16 @@
 #include <gdk/gdk.h>
 #include <gdk/gdkkeysyms.h>
 
+#ifdef GDK_WINDOWING_X11 
+#include <X11/X.h>
+#include <X11/Xlib.h>
+#include <X11/keysym.h>
+#include <gdk/gdkx.h>
+#endif
+
 #include "scim-bridge-client-key-event-utility-gtk.h"
 #include "scim-bridge-key-event.h"
 
-
 /* Helper Functions */
 static GdkKeymap *get_gdk_keymap (GdkWindow *window)
 {
@@ -96,4 +102,25 @@ void scim_bridge_key_event_gdk_to_bridge (ScimBridgeKeyEvent *bridge_key_event,
     } else {
         scim_bridge_key_event_set_pressed (bridge_key_event, FALSE);
     }
+    
+#ifdef GDK_WINDOWING_X11
+    Display *display = NULL;
+
+    if (window != NULL) {
+        display = GDK_WINDOW_XDISPLAY (window);
+    } else {
+        display = GDK_DISPLAY_XDISPLAY (gdk_display_get_default ());
+    }
+    
+    if (scim_bridge_key_event_get_code (bridge_key_event) == SCIM_BRIDGE_KEY_CODE_backslash) {
+        boolean kana_ro = FALSE;
+        int keysym_size = 0;
+        KeySym *keysyms = XGetKeyboardMapping (display, key_event->hardware_keycode, 1, &keysym_size);
+        if (keysyms != NULL) {
+            kana_ro = (keysyms[0] == XK_backslash && keysyms[1] == XK_underscore);
+            XFree (keysyms);
+        }
+        scim_bridge_key_event_set_quirk_enabled (bridge_key_event, SCIM_BRIDGE_KEY_QUIRK_KANA_RO, kana_ro);
+    }
+#endif
 }
