@@ -30,20 +30,14 @@
 #include "config.h"
 #endif
 
+#include "intl.h"
+
 #include <sys/types.h>
 
 #ifndef SCIM_VERSION
 #define SCIM_VERSION "<unknown>"
 #endif
 
-#define _(String)   gettext(String)
-#define N_(String)  gettext_noop(String)
-#define gettext_noop(String) (String)
-
-#ifndef NPOS
-#define NPOS -1
-#endif
-
 /**
  * The type of boolean.
  */
